/**
 **	File:	gclasslib.cpp
 **
 **	Published / author: 2004-04-14 / grymse@alhem.net
 **/

/*
Copyright (C) 2004  Anders Hedstrom

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

//int mkdir(const char *pathname, mode_t mode);


#include "sql2class.h"
#include "utils.h"

#include "gclasslib.h"


void generate_lib(short overwrite,char *dbname,short global,char *defdir,char *libname,char *prefix)
{
	FILE *fil;
	TBLS *t;
	time_t ti = time(NULL);
	struct tm *tp = localtime(&ti);
	char filename[200];
	char define[200];

// make .h file

	if (global)
		sprintf(filename,"%s/include/",prefix);
	else
		*filename = 0;
	sprintf(filename + strlen(filename),"%s.h",libname);
	sprintf(define,"_%s_H",libname);
	strupr(define);

	if ((fil = fopen(filename,"rt")) != NULL)
	{
		fclose(fil);
		if (!overwrite)
		{
			printf("'%s' already exists - not replaced\n",filename);
			return;
		}
	}
	if ((fil = fopen(filename,"wt")) == NULL)
	{
		printf("Couldn't create '%s'...\n",filename);
		return;
	}

	fprintf(fil,"/*\n");
	fprintf(fil," *\t %s\n",filename);
	fprintf(fil," *\t Generated by sql2class by (C) AH 2000\n");
	fprintf(fil," *\t %s",asctime(tp));
	fprintf(fil," */\n");
	fprintf(fil,"\n");

	fprintf(fil,"/*\n"
"Copyright (C) 2001  Anders Hedstrom (grymse@alhem.net)\n"
"\n"
"This program is free software; you can redistribute it and/or\n"
"modify it under the terms of the GNU General Public License\n"
"as published by the Free Software Foundation; either version 2\n"
"of the License, or (at your option) any later version.\n"
"\n"
"This program is distributed in the hope that it will be useful,\n"
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"GNU General Public License for more details.\n"
"\n"
"You should have received a copy of the GNU General Public License\n"
"along with this program; if not, write to the Free Software\n"
"Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n"
"*/\n");
	fprintf(fil,"\n");

	fprintf(fil,"#ifndef %s\n",define);
	fprintf(fil,"#define %s\n",define);
	fprintf(fil,"\n");
//	fprintf(fil,"#include <Query.h>\n");
//	fprintf(fil,"\n");
	fprintf(fil,"#ifdef _WIN32\n");
	fprintf(fil,"#define strncasecmp strnicmp\n");
	fprintf(fil,"#define strcasecmp stricmp\n");
	fprintf(fil,"#endif // _WIN32\n");
	if (container)
	{
		fprintf(fil,"#ifndef _LISTTYPE\n");
		fprintf(fil,"#define _LISTTYPE\n");
		fprintf(fil,"\ttypedef enum { Direct, InDirect } ListType;\n");
		fprintf(fil,"#endif\n");
		fprintf(fil,"\n");
	}

	if (*baseclass)
	{
		fprintf(fil,"#include \"%s.h\"\n\n",baseclass);
	}

	for (t = tblsbase; t; t = t -> next)
		generate_libh(fil,t);
	fprintf(fil,"#endif // %s\n",define);
	fclose(fil);


// make .cpp file

	if (global)
	{
		sprintf(filename,"%s/src/%s",prefix,libname);
		mkdir(filename,0755);
		strcat(filename,"/");
	} else
		*filename = 0;
	sprintf(filename + strlen(filename),"%s.cpp",libname);

	if ((fil = fopen(filename,"rt")) != NULL)
	{
		fclose(fil);
		if (!overwrite)
		{
			printf("'%s' already exists - not replaced\n",filename);
			return;
		}
	}
	if ((fil = fopen(filename,"wt")) == NULL)
	{
		printf("Couldn't create '%s'...\n",filename);
		return;
	}

	fprintf(fil,"/*\n");
	fprintf(fil," *\t %s\n",filename);
	fprintf(fil," *\t Generated by sql2class by (C) AH 2000\n");
	fprintf(fil," *\t %s",asctime(tp));
	fprintf(fil," */\n");
	fprintf(fil,"\n");

	fprintf(fil,"/*\n"
"Copyright (C) 2001  Anders Hedstrom (grymse@alhem.net)\n"
"\n"
"This program is free software; you can redistribute it and/or\n"
"modify it under the terms of the GNU General Public License\n"
"as published by the Free Software Foundation; either version 2\n"
"of the License, or (at your option) any later version.\n"
"\n"
"This program is distributed in the hope that it will be useful,\n"
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"GNU General Public License for more details.\n"
"\n"
"You should have received a copy of the GNU General Public License\n"
"along with this program; if not, write to the Free Software\n"
"Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n"
"*/\n");
	fprintf(fil,"\n");

	fprintf(fil,"#include <stdio.h>\n");
	fprintf(fil,"#include <stdlib.h>\n");
	fprintf(fil,"#include <string.h>\n");
	fprintf(fil,"#include <mysql/mysql.h>\n");
//	fprintf(fil,"#include <Query.h>\n");
	fprintf(fil,"#include <libsql++.h>\n");
	if (mediumtext)
	{
		fprintf(fil,"#include <vector>\n"
			"#include <string>\n"
			"\n");
	}
	if (global)
		fprintf(fil,"#include <%s.h>\n",libname);
	else
		fprintf(fil,"#include \"%s.h\"\n",libname);
	fprintf(fil,"\n");
	for (t = tblsbase; t; t = t -> next)
		generate_libc(fil,t);
	fclose(fil);


// make Makefile

	if (global)
	{
		sprintf(filename,"%s/src/%s/",prefix,libname);
		sprintf(filename + strlen(filename),"Makefile");

		if ((fil = fopen(filename,"rt")) != NULL)
		{
			fclose(fil);
			if (!overwrite)
			{
				printf("'%s' already exists - not replaced\n",filename);
				return;
			}
		}
		if ((fil = fopen(filename,"wt")) == NULL)
		{
			printf("Couldn't create '%s'...\n",filename);
			return;
		}
		fprintf(fil,"# makefile for '%s'\n",libname);
		fprintf(fil,"INCLUDE =	-I/usr/devel/include -I.\n");
		fprintf(fil,"CFLAGS =	-Wall -g $(INCLUDE)\n");
		fprintf(fil,"CPPFLAGS =	$(CFLAGS)\n");

		fprintf(fil,"PROGS =\t%s.a\n",libname);
		fprintf(fil,"\n");
		fprintf(fil,"all:\t$(PROGS)\n");
		fprintf(fil,"\n");
		fprintf(fil,"install:\tall\n");
		fprintf(fil,"\t\tcp %s.a ../../lib\n",libname);
		fprintf(fil,"\n");
		fprintf(fil,"LIB =\t%s.o\n",libname);
		fprintf(fil,"%s.a:\t$(LIB)\n",libname);
		fprintf(fil,"\t\tar cru %s.a $(LIB)\n",libname);
		fprintf(fil,"\t\tranlib %s.a\n",libname);
		fprintf(fil,"\n");
		fprintf(fil,"%s.o:\t%s.cpp ../../include/%s.h\n",libname,libname,libname);

		fclose(fil);
	} // only make Makefile if 'global'

}


void generate_libh(FILE *fil,TBLS *table)
{
	FIELDS *t;
	char *ptrtotable;
	char classname[200];
	char slask[200];
	char cclassname[200];	// container class name
	char cstruct[200];	// container struct
	char lwrclassname[200];
	char slask2[200];

	strcpy(classname,table -> name);
	strlwr(classname);
	strcpy(lwrclassname,classname);
	if (*classname >= 'a' && *classname <= 'z')
		*classname -= 32;

	sprintf(cclassname,"c%s",classname);	// container class name
	if (!strcmp(classname + strlen(classname) - 2,"ss"))
		sprintf(cstruct,"c%struct",classname);
	else
		sprintf(cstruct,"c%sstruct",classname); // container struct


	fprintf(fil,"\n");
	fprintf(fil,"/**\n");
	fprintf(fil," **  Class '%s' and '%s'\n",classname,cclassname);
	fprintf(fil," **/\n");
	fprintf(fil,"\n");

	if (*namespc)
	{
		fprintf(fil,"namespace %s\n{\n",namespc);
	}
	if (container)
	{
		fprintf(fil,"\ttypedef struct %s\n",cstruct);
		fprintf(fil,"\t{\n");
		fprintf(fil,"\t\tstruct %s *next;\n",cstruct);
		fprintf(fil,"\t\tclass %s *%s;\n",classname,lwrclassname);
		fprintf(fil,"\t};\n");
		fprintf(fil,"\n");
	}

	if (*baseclass)
	{
		fprintf(fil,"class %s : public %s {\n",classname,baseclass);
	}
	else
	{
		fprintf(fil,"class %s {\n",classname);
	}
	fprintf(fil,"public:\n");

// empty constructor
/*
	fprintf(fil,"\t%s();\n",classname);
*/

// empty constructor + database
	fprintf(fil,"\t%s(Database *);\n",classname);

// data fetch constructor + database
	fprintf(fil,"\t%s(Database *,long);\n",classname);

// data fetch constructor + database
	fprintf(fil,"\t%s(Database *,const std::string& );\n",classname);

// data fetch constructor from query
	fprintf(fil,"\t%s(Database *,Query *);\n",classname);

// destructor
	fprintf(fil,"\t~%s();\n",classname);


// void Class::select(const std::string& ) 'spawn_wrapper'
//  numera ersatt av Class::Class(db,sql)...
	fprintf(fil,"\tvoid select(const std::string& );\n");


// long Class::insert()
//  'insert into ...(...) values(...)'
	fprintf(fil,"\tlong insert();\n");


// void Class::update()
	fprintf(fil,"\tvoid update();\n");

// void Class::update(long num)
//  'update ... set ... where ...'
	fprintf(fil,"\tvoid update(long);\n");

// void Class::save()
	fprintf(fil,"\tvoid save();\n");

// void Class::erase()
	fprintf(fil,"\tvoid erase();\n");

// void Class::xml()
	fprintf(fil,"\tstd::string xml();\n");
	fprintf(fil,"\tstd::string xml(const std::string& ,const std::string& );\n");
	fprintf(fil,"\tint num_cols();\n");

// variables
	fprintf(fil,"\n");
	for (t = table -> fieldsbase; t; t = t -> next)
	{
		if (!strcmp(t -> ctype,"double"))
			fprintf(fil,"\t%s %s;",t -> ctype,t -> column);
		else
		if (!strcmp(t -> ctype,"long"))
			fprintf(fil,"\t%s %s;",t -> ctype,t -> column);
		else
		if (!strcmp(t -> ctype,"text"))
		{
			if (use_stl)
			{
				fprintf(fil,"\tstd::string %s;",t -> column);
			}
			else
			{
				fprintf(fil,"\tchar *%s;",t -> column);
			}
		}
		else
		if (!strcmp(t -> ctype,"mediumtext"))
			fprintf(fil,"\tstd::vector<std::string> %s;",t -> column);
		else
		if (*t -> ctype)
		{
			if (use_stl)
			{
				fprintf(fil,"\tstd::string %s;",t -> column);
			}
			else
			{
				fprintf(fil,"\tchar %s[%d];",t -> column,t -> length + 1);
			}
		}
		else
			fprintf(fil,"\t// %d - %s %s",t -> num,t -> column,t -> typ);
 		if (*t -> comment)
			fprintf(fil," //!< %s ",t -> comment);
		fprintf(fil,"\n");

		if (*t -> comment && (ptrtotable = strstr(t -> comment,"ptrtotable: ")) != NULL)
		{
			strcpy(slask,ptrtotable + 12);	// ptrtoclass
			strlwr(slask);
			strcpy(slask2,slask);
			if (*slask >= 'a' && *slask <= 'z')
				*slask -= 32;
			fprintf(fil,"\tclass %s *_%s;\n",slask,t -> column);
		}
	}
	fprintf(fil,"\n");


// private
	fprintf(fil,"private:\n");

	fprintf(fil,"\tDatabase *database;\n");
	fprintf(fil,"\tshort new_object;\n");

// clear fields
	fprintf(fil,"\tvoid clear();\n");

// spawn from query
	fprintf(fil,"\tvoid spawn(const std::string& );\n");

// spawn from query
	fprintf(fil,"\tvoid spawn(Query *);\n");

	fprintf(fil,"}; // End of class '%s'\n",classname);
	fprintf(fil,"\n");


/*
 * Create container class (see container.h for an example)
 */

	if (container)
	{
	fprintf(fil,"class %s {\n",cclassname);
	fprintf(fil,"public:\n");

	// constructor cclassname(db) - empty list
	fprintf(fil,"\t%s::%s(Database *,ListType);\n",cclassname,cclassname);

	// constructor cclassname(db,sql)
	fprintf(fil,"\t%s::%s(Database *,char *,ListType);\n",cclassname,cclassname);

	// destructor
	fprintf(fil,"\t%s::~%s();\n",cclassname,cclassname);

	// save()
	fprintf(fil,"\tvoid %s::save();\n",cclassname);

	// baseptr()
	fprintf(fil,"\t%s *%s::baseptr();\n",cstruct,cclassname);

	// get...()
	fprintf(fil,"\t%s *%s::get%s(long);\n",classname,cclassname,lwrclassname);

	// add(classname *)
	fprintf(fil,"\tvoid %s::add(%s *);\n",cclassname,classname);

	// remove(classname *)
	fprintf(fil,"\tvoid %s::remove(%s *);\n",cclassname,classname);

	// find_...()
	fprintf(fil,"\n");
	fprintf(fil,"\t// find_xxx()\n");
	for (t = table -> fieldsbase; t; t = t -> next)
	{
		strcpy(slask,typestring(t));
		if (*slask)
		{
			if (slask[strlen(slask) - 1] == ' ')
				slask[strlen(slask) - 1] = 0;
			fprintf(fil,"\t%s *%s::find_%s(%s);\n",classname,cclassname,t -> column,slask);
		}
	}

	// private
	fprintf(fil,"private:\n");
	fprintf(fil,"\tDatabase *database;\n");
	fprintf(fil,"\t%s *base;\n",cstruct);
	fprintf(fil,"\t%s **%ss;\n",classname,lwrclassname);
	fprintf(fil,"\tlong qty;\n");
	fprintf(fil,"\tListType _lt;\n");
	fprintf(fil,"}; // End of class '%s'\n",cclassname);
	fprintf(fil,"\n");
	} // if (container)

	if (*namespc)
	{
		fprintf(fil,"} // end of namespace\n");
	}
}


void generate_libc(FILE *fil,TBLS *table)
{
	FIELDS *t;
	char *ptrtotable;
	int i;
//	short first;
//	char c;
	char classname[200];
	char slask[200];
	char cclassname[200];	// container class name
	char cstruct[200];	// container struct
	char lwrclassname[200];
	char lwrtablename[200];

	strcpy(classname,table -> name);
	strlwr(classname);
	strcpy(lwrclassname,classname);
	if (*classname >= 'a' && *classname <= 'z')
		*classname -= 32;
		
	strcpy(lwrtablename,table -> name);
	strlwr(lwrtablename);
	if (*lwrtablename >= 'a' && *lwrtablename <= 'z')
		*lwrtablename -= 32;

	sprintf(cclassname,"c%s",lwrclassname);	// container class name
	cclassname[1] -= 32; // (uppercase)
	if (!strcmp(classname + strlen(classname) - 2,"ss"))
		sprintf(cstruct,"c%struct",classname);
	else
		sprintf(cstruct,"c%sstruct",classname); // container struct
	if (*namespc)
	{
		fprintf(fil,"namespace %s {\n", namespc);
	}
	fprintf(fil,"\n");
	fprintf(fil,"/**\n **  Begin class '%s'\n **/\n",classname);
	fprintf(fil,"\n");

//	fprintf(fil,"/* public: */\n");
//	fprintf(fil,"\n");

// empty constructor
/*
	fprintf(fil,"\t%s::%s()\n\t{\n",classname,lwrtablename);
	fprintf(fil,"\t\tdatabase = NULL;\n");
	fprintf(fil,"\t\tnew_object = 0;\n");
	fprintf(fil,"\t\tclear();\n");
	fprintf(fil,"\t}\n");
*/

// empty constructor + database
	fprintf(fil,"\t%s::%s(Database *db)\n\t{\n",classname,lwrtablename);
	fprintf(fil,"\t\tdatabase = db;\n");
	fprintf(fil,"\t\tnew_object = 1;\n");
	fprintf(fil,"\t\tclear();\n");
	fprintf(fil,"\t}\n");

// data fetch constructor + database
	fprintf(fil,"\t%s::%s(Database *db,long ix)\n\t{\n",classname,lwrtablename);
	fprintf(fil,"\t\tchar sql[80];\n");
	fprintf(fil,"\t\tdatabase = db;\n");
	fprintf(fil,"\t\tnew_object = 1;\n");
	fprintf(fil,"\t\tsprintf(sql,\"select * from %s where %s=%%ld\",ix);\n",table -> name,table -> fieldsbase -> column);
	// call spawn here
	fprintf(fil,"\t\tspawn(sql);\n");
	fprintf(fil,"\t}\n");

// data fetch constructor + database
	fprintf(fil,"\t%s::%s(Database *db,const std::string& sql)\n\t{\n",classname,lwrtablename);
	fprintf(fil,"\t\tdatabase = db;\n");
	fprintf(fil,"\t\tnew_object = 1;\n");
	// call spawn here
	fprintf(fil,"\t\tspawn(sql);\n");
	fprintf(fil,"\t}\n");

// data fetch constructor from query
	fprintf(fil,"\t%s::%s(Database *db,Query *qd)\n\t{\n",classname,lwrtablename);
	fprintf(fil,"\t\tdatabase = db;\n");
	fprintf(fil,"\t\tnew_object = 0;\n");
	// call spawn here
	fprintf(fil,"\t\tspawn(qd);\n");
	fprintf(fil,"\t}\n");

// destructor
	fprintf(fil,"\t%s::~%s()\n\t{\n",classname,lwrtablename);
	for (t = table -> fieldsbase; t; t = t -> next)
		if (!strcmp(t -> ctype,"text"))
		{
			if (!use_stl)
			{
				fprintf(fil,"\t\tif (%s)\n",t -> column);
				fprintf(fil,"\t\t\tdelete %s;\n",t -> column);
			}
		}
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");


// void Class::select(const std::string& ) 'spawn_wrapper'
//  numera ersatt av Class::Class(db,sql)...
	fprintf(fil,"\tvoid %s::select(const std::string& sql)\n\t{\n",classname);
	fprintf(fil,"\t\tspawn(sql);\n");
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");


// long Class::insert()
//  'insert into ...(...) values(...)'
	fprintf(fil,"\tlong %s::insert()\n\t{\n",classname);
	fprintf(fil,"\t\tQuery q(database);\n");
//	fprintf(fil,"\t\tchar *sql = new char[32000];\n");
	fprintf(fil,"\t\tstd::string sql;\n");
	fprintf(fil,"\n");

//	fprintf(fil,"\t\tsprintf(sql,\"insert into %s(",table -> name);
	fprintf(fil,"\t\tsql = \"insert into %s(",table -> name);
	*slask = 0;
	for (t = table -> fieldsbase; t; t = t -> next)
	if (!t -> ignore)
	{
		fprintf(fil,"%s%s",slask,t -> column);
		strcpy(slask,",");
	} // !ignore
//	fprintf(fil,")\"\n\t\t\t\" values(");
	fprintf(fil,")\";\n");

	strcpy(slask," values(");
	for (t = table -> fieldsbase; t; t = t -> next)
	if (!t -> ignore)
	{
		if (!strcmp(t -> ctype,"double"))
		{
//			fprintf(fil,"%s%%f",slask);
			fprintf(fil,"\t\t{\n");
			fprintf(fil,"\t\t\tchar slask[100];\n");
			fprintf(fil,"\t\t\tsprintf(slask,\"%s%%f\",%s);\n",slask,t -> column);
			fprintf(fil,"\t\t\tsql += slask;\n");
			fprintf(fil,"\t\t}\n");
		}
		else
		if (!strcmp(t -> ctype,"long"))
		{
//			fprintf(fil,"%s%%ld",slask);
			fprintf(fil,"\t\t{\n");
			fprintf(fil,"\t\t\tchar slask[100];\n");
			fprintf(fil,"\t\t\tsprintf(slask,\"%s%%ld\",%s);\n",slask,t -> column);
			fprintf(fil,"\t\t\tsql += slask;\n");
			fprintf(fil,"\t\t}\n");
		}
		else
		if (!strcmp(t -> ctype,"text"))
		{
//			fprintf(fil,"%s'%%s'",slask);
			fprintf(fil,"\t\tsql += \"%s'\" + q.safestr(%s) + \"'\";\n",slask,t -> column);
		}
		else
		if (*t -> ctype)
		{
//			fprintf(fil,"%s'%%s'",slask);
			fprintf(fil,"\t\tsql += \"%s'\" + q.safestr(%s) + \"'\";\n",slask,t -> column);
		}
		else
		{
//			fprintf(fil,"%s''",slask);	// unsupported data type
			fprintf(fil,"\t\tsql += \"%s''\";\n",slask);
		}
		strcpy(slask,", ");
	} // !ignore
	fprintf(fil,"\t\tsql += \")\";\n");

	fprintf(fil,"\t\tq.execute(sql);\n");
	fprintf(fil,"\t\tnew_object = 0;\n");

//	fprintf(fil,"\t\tdelete sql;\n");
	fprintf(fil,"\t\treturn %s = q.insert_id();\n",table -> fieldsbase -> column);
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");


// void Class::update()
	fprintf(fil,"\tvoid %s::update()\n\t{\n",classname);
	fprintf(fil,"\t\t%s::update(%s);\n",classname,table -> fieldsbase -> column);
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

// void Class::update(long num)
//  'update ... set ... where ...'
	fprintf(fil,"\tvoid %s::update(long ix)\n\t{\n",classname);
	fprintf(fil,"\t\tQuery q(database);\n");
//	fprintf(fil,"\t\tchar *sql = new char[32000];\n");
	fprintf(fil,"\t\tstd::string sql;\n");

	sprintf(slask,"update %s set ",table -> name);
	for (t = table -> fieldsbase -> next; t; t = t -> next)
	if (!t -> ignore)
	{
		if (!strcmp(t -> ctype,"double"))
		{
//			fprintf(fil,"%c%s=%%f",c,t -> column);
			fprintf(fil,"\t\t{\n");
			fprintf(fil,"\t\t\tchar slask[200];\n");
			fprintf(fil,"\t\t\tsprintf(slask,\"%s%s='%%f'\",%s);\n",slask,t -> column,t -> column);
			fprintf(fil,"\t\t\tsql += slask;\n");
			fprintf(fil,"\t\t}\n");
			strcpy(slask,", ");
		} else
		if (!strcmp(t -> ctype,"long"))
		{
//			fprintf(fil,"%c%s=%%ld",c,t -> column);
			fprintf(fil,"\t\t{\n");
			fprintf(fil,"\t\t\tchar slask[200];\n");
			fprintf(fil,"\t\t\tsprintf(slask,\"%s%s='%%ld'\",%s);\n",slask,t -> column,t -> column);
			fprintf(fil,"\t\t\tsql += slask;\n");
			fprintf(fil,"\t\t}\n");
			strcpy(slask,", ");
		} else
		if (*t -> ctype)
		{
//			fprintf(fil,"%c%s='%%s'",c,t -> column);
			fprintf(fil,"\t\tsql += \"%s%s='\" + q.safestr(%s) + \"'\";\n",slask,t -> column,t -> column);
			strcpy(slask,", ");
		}
	} // !ignore
	{
		fprintf(fil,"\t\t{\n");
		fprintf(fil,"\t\t\tchar slask[200];\n");
		fprintf(fil,"\t\t\tsprintf(slask,\" where %s='%%ld'\",ix);\n",table -> fieldsbase -> column);
		fprintf(fil,"\t\t\tsql += slask;\n");
		fprintf(fil,"\t\t}\n");
	}

	fprintf(fil,"\t\tq.execute(sql);\n");

//	fprintf(fil,"\t\tdelete sql;\n");
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

// void Class::save()
	fprintf(fil,"\tvoid %s::save()\n\t{\n",classname);
	fprintf(fil,"\t\tif (new_object)\n");
	fprintf(fil,"\t\t\tinsert();\n");
	fprintf(fil,"\t\telse\n");
	fprintf(fil,"\t\t\tupdate();\n");
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

// void Class::erase()
	fprintf(fil,"\tvoid %s::erase()\n\t{\n",classname);
	fprintf(fil,"\t\tQuery q(database);\n");
	fprintf(fil,"\t\tchar sql[200];\n");
	fprintf(fil,"\t\tsprintf(sql,\"delete from %s where %s=%%ld\",%s);\n",
		table -> name,table -> fieldsbase -> column,table -> fieldsbase -> column);
	fprintf(fil,"\t\tq.execute(sql);\n");
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

// void Class::xml()
	fprintf(fil,"\tstd::string %s::xml()\n\t{\n",classname);
	fprintf(fil,"\t\tQuery q(database);\n");
//	fprintf(fil,"\t\t*dest = 0;\n");
	fprintf(fil,"\t\tstd::string dest;\n");
	fprintf(fil,"\t\tchar slask[200];\n");

	strcpy(slask,classname);
	strupr(slask);
//	fprintf(fil,"\t\tstrcpy(dest,\"<%s>\");\n",slask);
	fprintf(fil,"\t\tdest = \"<%s>\";\n",slask);

	for (t = table -> fieldsbase; t; t = t -> next)
	{
		strcpy(slask,t -> column);
		strupr(slask);
		if (!strcmp(t -> ctype,"double"))
		{
			fprintf(fil,"\t\tsprintf(slask,\"<%s>%%f</%s>\",%s);\n",slask,slask,t -> column);
			fprintf(fil,"\t\tdest += slask;\n");
		}
		else
		if (!strcmp(t -> ctype,"long"))
		{
			fprintf(fil,"\t\tsprintf(slask,\"<%s>%%ld</%s>\",%s);\n",slask,slask,t -> column);
			fprintf(fil,"\t\tdest += slask;\n");
		}
		else
		if (!strcmp(t -> ctype,"text"))
		{
//			fprintf(fil,"\t\tsprintf(dest + strlen(dest),\"<%s>%%s</%s>\",q.xmlsafestr(%s).c_str());\n",slask,slask,t -> column);
			fprintf(fil,"\t\tdest += \"<%s>\" + q.xmlsafestr(%s) + \"</%s>\";\n",slask,t -> column,slask);
		}
		else
		if (!strcmp(t -> ctype,"mediumtext"))
		{
			// vector<string>
		}
		else
		if (*t -> ctype)
		{
//			fprintf(fil,"\t\tsprintf(dest + strlen(dest),\"<%s>%%s</%s>\",q.xmlsafestr(%s).c_str());\n",slask,slask,t -> column);
			fprintf(fil,"\t\tdest += \"<%s>\" + q.xmlsafestr(%s) + \"</%s>\";\n",slask,t -> column,slask);
		}
	}
	strcpy(slask,classname);
	strupr(slask);
//	fprintf(fil,"\t\tstrcat(dest,\"</%s>\");\n",slask);
	fprintf(fil,"\t\tdest += \"</%s>\";\n",slask);

	fprintf(fil,"\t\treturn dest;\n");
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

// void Class::xml()
	fprintf(fil,"\tstd::string %s::xml(const std::string& tag,const std::string& xvalx)\n\t{\n",classname);
	fprintf(fil,"\t\tQuery q(database);\n");
//	fprintf(fil,"\t\t*dest = 0;\n");
	fprintf(fil,"\t\tstd::string dest;\n");
	fprintf(fil,"\t\tchar slask[200];\n");

	strcpy(slask,classname);
	strupr(slask);
//	fprintf(fil,"\t\tsprintf(dest,\"<%s %%s=\\\"%%s\\\">\",tag.c_str(),q.xmlsafestr(xvalx).c_str());\n",slask);
	fprintf(fil,"\t\tdest = \"<%s \" + tag + \"=\\\"\" + xvalx + \"\\\">\";\n",slask);

	for (t = table -> fieldsbase; t; t = t -> next)
	{
		strcpy(slask,t -> column);
		strupr(slask);
		if (!strcmp(t -> ctype,"double"))
		{
			fprintf(fil,"\t\tsprintf(slask,\"<%s>%%f</%s>\",%s);\n",slask,slask,t -> column);
			fprintf(fil,"\t\tdest += slask;\n");
		}
		else
		if (!strcmp(t -> ctype,"long"))
		{
			fprintf(fil,"\t\tsprintf(slask,\"<%s>%%ld</%s>\",%s);\n",slask,slask,t -> column);
			fprintf(fil,"\t\tdest += slask;\n");
		}
		else
		if (!strcmp(t -> ctype,"text"))
		{
//			fprintf(fil,"\t\tsprintf(dest + strlen(dest),\"<%s>%%s</%s>\",q.xmlsafestr(%s).c_str());\n",slask,slask,t -> column);
			fprintf(fil,"\t\tdest += \"<%s>\" + q.xmlsafestr(%s) + \"</%s>\";\n",slask,t -> column,slask);
		}
		else
		if (!strcmp(t -> ctype,"mediumtext"))
		{
			// vector<string>
		}
		else
		if (*t -> ctype)
		{
//			fprintf(fil,"\t\tsprintf(dest + strlen(dest),\"<%s>%%s</%s>\",q.xmlsafestr(%s).c_str());\n",slask,slask,t -> column);
			fprintf(fil,"\t\tdest += \"<%s>\" + q.xmlsafestr(%s) + \"</%s>\";\n",slask,t -> column,slask);
		}
	}
	strcpy(slask,classname);
	strupr(slask);
//	fprintf(fil,"\t\tstrcat(dest,\"</%s>\");\n",slask);
	fprintf(fil,"\t\tdest += \"</%s>\";\n",slask);

	fprintf(fil,"\t\treturn dest;\n");
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

// int Class::num_cols()
	int num_cols = 0;
	for (t = table -> fieldsbase; t; t = t -> next)
	{
		num_cols++;
	}
	fprintf(fil,"\tint %s::num_cols()\n",classname);
	fprintf(fil,"\t{\n");
	fprintf(fil,"\t\treturn %d;\n",num_cols); //listlen(table -> fieldsbase));
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");


// private
//	fprintf(fil,"/* private: */\n");
//	fprintf(fil,"\n");

// clear fields
	fprintf(fil,"\tvoid %s::clear()\n\t{\n",classname);
	for (t = table -> fieldsbase; t; t = t -> next)
	{
		if (!strcmp(t -> ctype,"double"))
			fprintf(fil,"\t\t%s = 0;\n",t -> column);
		else
		if (!strcmp(t -> ctype,"long"))
			fprintf(fil,"\t\t%s = 0;\n",t -> column);
		else
		if (!strcmp(t -> ctype,"text"))
		{
			if (use_stl)
			{
//				fprintf(fil,"\t\t%s = "";\n",t -> column);
			}
			else
			{
				fprintf(fil,"\t\t%s = NULL;\n",t -> column);
			}
		}
		else
		if (!strcmp(t -> ctype,"mediumtext"))
		{
			// a vector<string> is created empty
		}
		else
		if (*t -> ctype)
		{
			if (use_stl)
			{
//				fprintf(fil,"\t\t%s = "";\n",t -> column);
			}
			else
			{
				fprintf(fil,"\t\t*%s = 0;\n",t -> column);
				if (t -> length == 1) // char
					fprintf(fil,"\t\t%s[1] = 0;\n",t -> column);
			}
		} else
			fprintf(fil,"\t\t// %d - %s %s\n",t -> num,t -> column,t -> typ);

		if (*t -> comment && (ptrtotable = strstr(t -> comment,"ptrtotable: ")) != NULL)
		{
			strcpy(slask,ptrtotable + 12);	// ptrtoclass
			strlwr(slask);
			fprintf(fil,"\t\t_%s = NULL;\n",t -> column);
		}
	}
	fprintf(fil,"\t}\n");

// spawn from query
	fprintf(fil,"\tvoid %s::spawn(const std::string& sql)\n\t{\n",classname);
	fprintf(fil,"\t\tQuery q(database);\n");
//	fprintf(fil,"\t\tchar *temp = new char[32000];\n");
	fprintf(fil,"\t\tstd::string temp;\n");
	fprintf(fil,"\n");

	fprintf(fil,"\t\tif (!strncasecmp(sql.c_str(),\"select * \",9))\n");
	fprintf(fil,"\t\t{\n");
//	fprintf(fil,"\t\t\tsprintf(temp,\"");
	fprintf(fil,"\t\t\ttemp = \"");
	strcpy(slask,"select ");
	for (t = table -> fieldsbase; t; t = t -> next)
	{
		fprintf(fil,"%s%s",slask,t -> column);
		strcpy(slask,",");
	}
//	fprintf(fil," %%s\",sql.substr(9) );\n");
	fprintf(fil," \" + sql.substr(9);\n");
	fprintf(fil,"\t\t} else\n");
//	fprintf(fil,"\t\t\tstrcpy(temp,sql.c_str() );\n");
	fprintf(fil,"\t\t\ttemp = sql;\n");

	fprintf(fil,"\t\tq.get_result(temp);\n");
	fprintf(fil,"\t\tif (q.fetch_row())\n");
	fprintf(fil,"\t\t{\n");
	for (t = table -> fieldsbase; t; t = t -> next)
	{
		if (*t -> ctype)
		{
			if (!strcmp(t -> ctype,"double"))
			{
				fprintf(fil,"\t\t\t%s = q.getnum(%d);",t -> column,t -> num);
				sprintf(slask,"%d",t -> num);
				i = strlen(t -> column) + strlen(slask);
			} else
			if (!strcmp(t -> ctype,"long"))
			{
				fprintf(fil,"\t\t\t%s = q.getval(%d);",t -> column,t -> num);
				sprintf(slask,"%d",t -> num);
				i = strlen(t -> column) + strlen(slask);
			} else
			if (!strcmp(t -> ctype,"text"))
			{
				if (use_stl)
				{
					fprintf(fil,"\t\t\t%s = q.getstr(%d);",t -> column,t -> num);
					sprintf(slask,"%d",t -> num);
					i = strlen(t -> column) + strlen(slask);
				}
				else
				{
					fprintf(fil,"\t\t\t%s = new char[strlen(q.getstr(%d)) + 1000];\n",t -> column,t -> num);
					fprintf(fil,"\t\t\tstrcpy(%s,q.getstr(%d));",t -> column,t -> num);
					sprintf(slask,"%d",t -> num);
					i = strlen(t -> column) + strlen(slask) + 6;
				}
			} else
			if (!strcmp(t -> ctype,"mediumtext"))
			{
				fprintf(fil,"\t\t\t{\n");
				fprintf(fil,"\t\t\t\tsize_t x = 0;\n");
				fprintf(fil,"\t\t\t\tchar *s = q.getstr(%d);\n",t -> num);
				fprintf(fil,"\t\t\t\tsize_t l = strlen(s);\n");
				fprintf(fil,"\t\t\t\tchar tmp[10000];\n");
				fprintf(fil,"\t\t\t\tfor (size_t i = 0; i < l; i++)\n");
				fprintf(fil,"\t\t\t\t{\n");
				fprintf(fil,"\t\t\t\t\tif (s[i] == '\\n')\n");
				fprintf(fil,"\t\t\t\t\t{\n");
				fprintf(fil,"\t\t\t\t\t\ttmp[x] = 0;\n");
				fprintf(fil,"\t\t\t\t\t\t%s.push_back(tmp);\n",t -> column);
				fprintf(fil,"\t\t\t\t\t\tx = 0;\n");
				fprintf(fil,"\t\t\t\t\t}\n");
				fprintf(fil,"\t\t\t\t\telse\n");
				fprintf(fil,"\t\t\t\t\t{\n");
				fprintf(fil,"\t\t\t\t\t\ttmp[x++] = s[i];\n");
				fprintf(fil,"\t\t\t\t\t}\n");
				fprintf(fil,"\t\t\t\t}\n");
				fprintf(fil,"\t\t\t}\n");
				//
				i = 0;
			}
			else
			{
				if (use_stl)
				{
					fprintf(fil,"\t\t\t%s = q.getstr(%d);",t -> column,t -> num);
					sprintf(slask,"%d",t -> num);
					i = strlen(t -> column) + strlen(slask);
				}
				else
				{
					fprintf(fil,"\t\t\tstrcpy(%s,q.getstr(%d));",t -> column,t -> num);
					sprintf(slask,"%d",t -> num);
					i = strlen(t -> column) + strlen(slask) + 6;
				}
			}
			if (i)
			{
				strcpy(slask,"                                        ");
				slask[40 - i] = 0;
				fprintf(fil,"%s// %d - %s %s\n",slask,t -> num,t -> column,t -> typ);
			}
		} else
		{
			fprintf(fil,"\t\t\t// %d - %s %s\n",t -> num,t -> column,t -> typ);
		}
	}
	fprintf(fil,"\t\t\tnew_object = 0;\n");
	fprintf(fil,"\t\t} else\n");
	fprintf(fil,"\t\t\tclear();\n");
	fprintf(fil,"\t\tq.free_result();\n");

//	fprintf(fil,"\t\tdelete temp;\n");
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

// spawn from query
	fprintf(fil,"\tvoid %s::spawn(Query *qd)\n",classname);
	fprintf(fil,"\t{\n");
	for (t = table -> fieldsbase; t; t = t -> next)
	{
		if (*t -> ctype)
		{
			if (!strcmp(t -> ctype,"double"))
			{
				fprintf(fil,"\t\t%s = qd -> getnum(%d);",t -> column,t -> num);
				sprintf(slask,"%d",t -> num);
				i = strlen(t -> column) + strlen(slask);
			} else
			if (!strcmp(t -> ctype,"long"))
			{
				fprintf(fil,"\t\t%s = qd -> getval(%d);",t -> column,t -> num);
				sprintf(slask,"%d",t -> num);
				i = strlen(t -> column) + strlen(slask);
			} else
			if (!strcmp(t -> ctype,"text"))
			{
				if (use_stl)
				{
					fprintf(fil,"\t\t%s = qd -> getstr(%d);",t -> column,t -> num);
					sprintf(slask,"%d",t -> num);
					i = strlen(t -> column) + strlen(slask);
				}
				else
				{
					fprintf(fil,"\t\t%s = new char[strlen(qd -> getstr(%d)) + 1000];\n",t -> column,t -> num);
					fprintf(fil,"\t\tstrcpy(%s,qd -> getstr(%d));",t -> column,t -> num);
					sprintf(slask,"%d",t -> num);
					i = strlen(t -> column) + strlen(slask) + 6;
				}
			} else
			if (!strcmp(t -> ctype,"mediumtext"))
			{
				fprintf(fil,"\t\t{\n");
				fprintf(fil,"\t\t\tsize_t x = 0;\n");
				fprintf(fil,"\t\t\tchar *s = qd -> getstr(%d);\n",t -> num);
				fprintf(fil,"\t\t\tsize_t l = strlen(s);\n");
				fprintf(fil,"\t\t\tchar tmp[10000];\n");
				fprintf(fil,"\t\t\tfor (size_t i = 0; i < l; i++)\n");
				fprintf(fil,"\t\t\t{\n");
				fprintf(fil,"\t\t\t\tif (s[i] == '\\n')\n");
				fprintf(fil,"\t\t\t\t{\n");
				fprintf(fil,"\t\t\t\t\ttmp[x] = 0;\n");
				fprintf(fil,"\t\t\t\t\t%s.push_back(tmp);\n",t -> column);
				fprintf(fil,"\t\t\t\t\tx = 0;\n");
				fprintf(fil,"\t\t\t\t}\n");
				fprintf(fil,"\t\t\t\telse\n");
				fprintf(fil,"\t\t\t\t{\n");
				fprintf(fil,"\t\t\t\t\ttmp[x++] = s[i];\n");
				fprintf(fil,"\t\t\t\t}\n");
				fprintf(fil,"\t\t\t}\n");
				fprintf(fil,"\t\t}\n");
				//
				i = 0;
			}
			else
			{
				if (use_stl)
				{
					fprintf(fil,"\t\t%s = qd -> getstr(%d);",t -> column,t -> num);
					sprintf(slask,"%d",t -> num);
					i = strlen(t -> column) + strlen(slask);
				}
				else
				{
					fprintf(fil,"\t\tstrcpy(%s,qd -> getstr(%d));",t -> column,t -> num);
					sprintf(slask,"%d",t -> num);
					i = strlen(t -> column) + strlen(slask) + 6;
				}
			}
			if (i)
			{
				strcpy(slask,"                                        ");
				slask[40 - i] = 0;
				fprintf(fil,"%s// %d - %s %s\n",slask,t -> num,t -> column,t -> typ);
			}
		} else
		{
			fprintf(fil,"\t\t// %d - %s %s\n",t -> num,t -> column,t -> typ);
		}
	}
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

	fprintf(fil,"// End of implementation of class '%s'\n",classname);
	fprintf(fil,"\n");

/*
 * Create container class (see container.h for an example)
 */

	if (container)
	{
	fprintf(fil,"\n");
	fprintf(fil,"/**\n **  Begin class '%s'\n **/\n",cclassname);
	fprintf(fil,"\n");

//	fprintf(fil,"/* public: */\n");
//	fprintf(fil,"\n");

	// constructor cclassname(db) - empty list
	fprintf(fil,"\t%s::%s(Database *db,ListType lt)\n\t{\n",cclassname,cclassname);
	fprintf(fil,"\t\tdatabase = db;\n");
	fprintf(fil,"\t\tbase = NULL;\n");
	fprintf(fil,"\t\t%ss = NULL;\n",lwrclassname);
	fprintf(fil,"\t\tqty = 0;\n");
	fprintf(fil,"\t\t_lt = lt;\n");
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

	// constructor cclassname(db,sql)
	fprintf(fil,"\t%s::%s(Database *db,char *sql,ListType lt)\n\t{\n",cclassname,cclassname);
	fprintf(fil,"\t\tQuery q(db);\n");
	fprintf(fil,"\t\t%s *item,*p;\n",cstruct);
	fprintf(fil,"\t\tlong l;\n");
	fprintf(fil,"\n");
	fprintf(fil,"\t\tdatabase = db;\n");
	fprintf(fil,"\t\tbase = NULL;\n");
	fprintf(fil,"\t\t%ss = NULL;\n",lwrclassname);
	fprintf(fil,"\t\tqty = 0;\n");
	fprintf(fil,"\t\t_lt = lt;\n");
	fprintf(fil,"\n");
	fprintf(fil,"\t\tq.get_result(sql);\n");
	fprintf(fil,"\t\twhile (q.fetch_row())\n");
	fprintf(fil,"\t\t{\n");
	fprintf(fil,"\t\t\titem = new %s;\n",cstruct);
	fprintf(fil,"\t\t\titem -> next = NULL;\n");
	fprintf(fil,"\t\t\titem -> %s = new %s(db,q.getval());\n",lwrclassname,classname);
	fprintf(fil,"\t\t\tif (!base)\n");
	fprintf(fil,"\t\t\t\tbase = item;\n");
	fprintf(fil,"\t\t\telse\n");
	fprintf(fil,"\t\t\t{\n");
	fprintf(fil,"\t\t\t\tp = base;\n");
	fprintf(fil,"\t\t\t\twhile (p -> next)\n");
	fprintf(fil,"\t\t\t\t\tp = p -> next;\n");
	fprintf(fil,"\t\t\t\tp -> next = item;\n");
	fprintf(fil,"\t\t\t}\n");
	fprintf(fil,"\t\t\tqty++;\n");
	fprintf(fil,"\t\t}\n");
	fprintf(fil,"\t\tq.free_result();\n");
	fprintf(fil,"\n");
	fprintf(fil,"\t\t%ss = new %s *[qty];\n",lwrclassname,classname);
	fprintf(fil,"\t\tl = 0;\n");
	fprintf(fil,"\t\tfor (p = base; p; p = p -> next)\n");
	fprintf(fil,"\t\t\t%ss[l++] = p -> %s;\n",lwrclassname,lwrclassname);
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

	// destructor
	fprintf(fil,"\t%s::~%s()\n\t{\n",cclassname,cclassname);
	fprintf(fil,"\t\t%s *q,*tmp;\n",cstruct);
	fprintf(fil,"\t\tfor (q = base; q; q = tmp)\n");
	fprintf(fil,"\t\t{\n");
	fprintf(fil,"\t\t\tif (_lt == Direct)\n");
	fprintf(fil,"\t\t\t\tdelete q -> %s;\n",lwrclassname);
	fprintf(fil,"\t\t\ttmp = q -> next;\n");
	fprintf(fil,"\t\t\tdelete q;\n");
	fprintf(fil,"\t\t}\n");
	fprintf(fil,"\t\tdelete %ss;\n",lwrclassname);
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

	// save()
	fprintf(fil,"\tvoid %s::save()\n\t{\n",cclassname);
	fprintf(fil,"\t\t%s *q;\n",cstruct);
	fprintf(fil,"\t\tfor (q = base; q; q = q -> next)\n");
	fprintf(fil,"\t\t\tq -> %s -> save();\n",lwrclassname);
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

	// baseptr()
	fprintf(fil,"\t%s *%s::baseptr()\n\t{\n",cstruct,cclassname);
	fprintf(fil,"\t\treturn base;\n");
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

	// get...()
	fprintf(fil,"\t%s *%s::get%s(long ix)\n\t{\n",classname,cclassname,lwrclassname);
	fprintf(fil,"\t\treturn %ss[ix];\n",lwrclassname);
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

	// add(classname *)
	fprintf(fil,"\tvoid %s::add(%s *ix)\n\t{\n",cclassname,classname);
	fprintf(fil,"\t\t%s *p,*item = new %s;\n",cstruct,cstruct);
	fprintf(fil,"\t\tlong l;\n");
	fprintf(fil,"\n");
	fprintf(fil,"\t\titem -> next = NULL;\n");
	fprintf(fil,"\t\titem -> %s = ix;\n",lwrclassname);
	fprintf(fil,"\t\tif (!base)\n");
	fprintf(fil,"\t\t\tbase = item;\n");
	fprintf(fil,"\t\telse\n");
	fprintf(fil,"\t\t{\n");
	fprintf(fil,"\t\t\tp = base;\n");
	fprintf(fil,"\t\t\twhile (p -> next)\n");
	fprintf(fil,"\t\t\t\tp = p -> next;\n");
	fprintf(fil,"\t\t\tp -> next = item;\n");
	fprintf(fil,"\t\t}\n");
	fprintf(fil,"\t\tqty++;\n");
	fprintf(fil,"\t\tdelete %ss;\n",lwrclassname);

	fprintf(fil,"\t\t%ss = new %s *[qty];\n",lwrclassname,classname);
	fprintf(fil,"\t\tl = 0;\n");
	fprintf(fil,"\t\tfor (p = base; p; p = p -> next)\n");
	fprintf(fil,"\t\t\t%ss[l++] = p -> %s;\n",lwrclassname,lwrclassname);

	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

	// remove(classname *)
	fprintf(fil,"\tvoid %s::remove(%s *ix)\n",cclassname,classname);
	fprintf(fil,"\t{\n");
	fprintf(fil,"\t\t%s *p,*q = NULL;\n",cstruct);
	fprintf(fil,"\t\tlong l;\n");
	fprintf(fil,"\t\tfor (p = base; p; p = p -> next)\n");
	fprintf(fil,"\t\t\tif (p -> %s == ix)\n",lwrclassname);
	fprintf(fil,"\t\t\t{\n");
	fprintf(fil,"\t\t\t\tif (q)\n");
	fprintf(fil,"\t\t\t\t\tq -> next = p -> next;\n");
	fprintf(fil,"\t\t\t\telse\n");
	fprintf(fil,"\t\t\t\t\tbase = p -> next;\n");
	fprintf(fil,"\t\t\t\tbreak;\n");
	fprintf(fil,"\t\t\t} else\n");
	fprintf(fil,"\t\t\t\tq = p;\n");
	fprintf(fil,"\t\tif (p)\n");
	fprintf(fil,"\t\t{\n");
	fprintf(fil,"\t\t\tdelete p;\n");
	fprintf(fil,"\t\t\tqty--;\n");

	fprintf(fil,"\t\t\tdelete %ss;\n",lwrclassname);

	fprintf(fil,"\t\t\t%ss = new %s *[qty];\n",lwrclassname,classname);
	fprintf(fil,"\t\t\tl = 0;\n");
	fprintf(fil,"\t\t\tfor (p = base; p; p = p -> next)\n");
	fprintf(fil,"\t\t\t\t%ss[l++] = p -> %s;\n",lwrclassname,lwrclassname);

	fprintf(fil,"\t\t}\n");
	fprintf(fil,"\t}\n");
	fprintf(fil,"\n");

	// find_...()
	fprintf(fil,"\t// find_xxx()\n\n");

	for (t = table -> fieldsbase; t; t = t -> next)
	{
		strcpy(slask,typestring(t));
		if (*slask)
		{
			fprintf(fil,"\t%s *%s::find_%s(%six)\n\t{\n",classname,cclassname,t -> column,slask);
			fprintf(fil,"\t\t%s *q;\n",cstruct);
			fprintf(fil,"\t\tfor (q = base; q; q = q -> next)\n");
			if (!strcmp(t -> ctype,"long") || !strcmp(t -> ctype,"double"))
				fprintf(fil,"\t\t\tif (q -> %s -> %s == ix)\n",lwrclassname,t -> column);
			else
				fprintf(fil,"\t\t\tif (!strcasecmp(q -> %s -> %s,ix))\n",lwrclassname,t -> column);
			fprintf(fil,"\t\t\t\tbreak;\n");
			fprintf(fil,"\t\treturn q ? q -> %s : NULL;\n",lwrclassname);
			fprintf(fil,"\t}\n");
			fprintf(fil,"\n");
		}
	}

	// private
//	fprintf(fil,"/* private: */\n");
//	fprintf(fil,"\n");

	fprintf(fil,"// End of implementation of class '%s'\n",cclassname);
	fprintf(fil,"\n");
	} // if (container)

	if (*namespc)
	{
		fprintf(fil,"} // End of namespace\n");
	}
}


